#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "log.h"
#include "fifo.h"
#include "socklib.h"
#include "udpcast.h"
#include "udpc-protoc.h"
#include "udp-sender.h"
#include "rate-limit.h"
#include "participants.h"
#include "statistics.h"

static time_t maxSendWait=0;
static time_t minSendWait=0;

/**
 * This file contains the code to set up the connection
 */

static int doTransfer(int socket, 
		      participantsDb_t db,
		      struct disk_config *disk_config,
		      struct net_config *net_config);


static int isPointToPoint(participantsDb_t db, int flags) {
    if(flags & FLAG_POINTOPOINT)
	return 1;
    if(flags & (FLAG_NOPOINTOPOINT | FLAG_ASYNC))
	return 0;
    return nrParticipants(db) == 1;
}


static int sendConnectionReply(participantsDb_t db,
			       int socket,
			       struct net_config *config,
			       struct sockaddr *client, 
			       int capabilities,
			       unsigned int rcvbuf) {
    struct connectReply reply;

    if(rcvbuf == 0)
	rcvbuf = 65536;

    if(capabilities & CAP_BIG_ENDIAN) {
	reply.opCode = htons(CMD_CONNECT_REPLY);
	reply.clNr = htonl(addParticipant(db,
					  client, 
					  capabilities,
					  rcvbuf,
					  config->flags & FLAG_POINTOPOINT));
	reply.blockSize = htonl(config->blockSize);
    } else if(capabilities & CAP_LITTLE_ENDIAN) {
	reply.opCode = htopcs(CMD_CONNECT_REPLY);
	reply.clNr = htopcl(addParticipant(db,
					   client, 
					   capabilities,
					   rcvbuf,
					   config->flags & FLAG_POINTOPOINT));
	reply.blockSize = htopcl(config->blockSize);
    }
    reply.reserved = 0;

    if(config->flags & FLAG_POINTOPOINT) {
	copyIpFrom(&config->dataMcastAddr, client);
    }

    /* new parameters: always big endian */
    reply.capabilities = ntohl(config->capabilities);
    copyToMessage(reply.mcastAddr,&config->dataMcastAddr);
    /*reply.mcastAddress = mcastAddress;*/
    doRateLimit(config->rateLimit, sizeof(reply));
    if(SEND(socket, reply, *client) < 0) {
	perror("reply add new client");
	return -1;
    }
    return 0;
}

static void sendHello(struct net_config *net_config, int socket) {
    struct hello hello;
    /* send hello message */
    hello.opCode = htopcs(CMD_HELLO);
    hello.reserved = 0;
    hello.capabilities = htonl(net_config->capabilities);
    copyToMessage(hello.mcastAddr,&net_config->dataMcastAddr);
    hello.blockSize = htons(net_config->blockSize);
    doRateLimit(net_config->rateLimit, sizeof(hello));
    BCAST_CONTROL(socket, hello);
}


/* *****************************************************
 * Receive and process a localization enquiry by a client
 */
static int mainDispatcher(int fd, 
			  participantsDb_t db,
			  struct disk_config *disk_config,
			  struct net_config *net_config,
			  int keyboard, int *tries)
{
    struct sockaddr client;
    union message fromClient;
    fd_set read_set;
    int ret;
    int msgLength;

    if (nrParticipants(db)){
	time_t now = time(0);
	flprintf("Ready. Press any key to start sending data.\n");
 	if(net_config->max_client_wait && (maxSendWait < 1)){
		maxSendWait = now + net_config->max_client_wait;
	}
 	if(net_config->min_client_wait && (minSendWait < 1)){
		minSendWait = now + net_config->min_client_wait;
	}
    } else {
	maxSendWait = 0;
    }
 
    while(1) {
	struct timeval tv;
	struct timeval *tvp;
	int nr_desc;

	FD_ZERO(&read_set);
	FD_SET(0, &read_set);
	FD_SET(fd, &read_set);

	if(net_config->rexmit_hello_interval) {
	    tv.tv_usec = (net_config->rexmit_hello_interval % 1000)*1000;
	    tv.tv_sec = net_config->rexmit_hello_interval / 1000;
	    tvp = &tv;
	} else if(nrParticipants(db)) {
	    tv.tv_usec = 0;
	    tv.tv_sec = 2;
	    tvp = &tv;
	} else {
	    tvp = 0;
	}
	nr_desc = select(fd+1,  &read_set, 0, 0,  tvp);
	if(nr_desc < 0) {
	    perror("%d select");
	    exit(1);
	}
	if(nr_desc == 0) {
	    sendHello(net_config, fd);
	    (*tries)++;
	    if(net_config->autostart != 0 && *tries > net_config->autostart) {
		doTransfer(fd, db, disk_config, net_config);
		return 1;
	    }
	} else
	    break;
	if (nrParticipants(db)){
		time_t now = time(0);
		/*
		 * Bad if statement...
		 *  If the maxSendWait time has passed transfer the data
		 */
		if(net_config->max_client_wait && (now >= maxSendWait)){
			doTransfer(fd, db, disk_config, net_config);
			return 1;
			/*
			 * Otherwise check to see if the minimum of clients
			 *  have checked in.
			 */
		} else if ((nrParticipants(db) >= net_config->min_clients)){
	 		/*
			 *  If they have and there's a min wait time, we'll 
			 *    wait around anyway until then.
			 */
			if(net_config->min_client_wait){
			    if(now >= minSendWait){
				doTransfer(fd, db, disk_config, net_config);
				return 1;
    }
			} else {
			/*
			 * Otherwise we transver the data.
			 */
			    doTransfer(fd, db, disk_config, net_config);
			    return 1;
			}
		}
	}
    }
	

    if(keyboard) {
	flprintf("KBD-ing.\n");
	if(FD_ISSET(0, &read_set)) {
	    char line[80];
	    read(0, line, 80);
	    if (line[0] == 'q') {
		exit(0);
	    }
	    doTransfer(fd, db, disk_config, net_config);
	    return 1;
	}
    }

    BZERO(fromClient); /* Zero it out in order to cope with short messages
			* from older versions */

    msgLength = RECV(fd, fromClient, client, net_config->portBase);
    if(msgLength < 0) {
	perror("problem getting data from client");
	return -1;
    }

    if(net_config->flags & FLAG_ASYNC)
	return 0;

	flprintf("Checking Connections.\n");
    switch(ntohs(fromClient.opCode)) {
	case CMD_CONNECT_REQ:
	    sendConnectionReply(db, fd,
				net_config,
				&client, 
				CAP_BIG_ENDIAN |
				ntohl(fromClient.connectReq.capabilities),
				ntohl(fromClient.connectReq.rcvbuf));
	    return 0;
	case CMD_GO:
	    doTransfer(fd, db, disk_config, net_config);
	    return 1;
	case CMD_DISCONNECT:
	    ret = lookupParticipant(db, &client);
	    if (ret >= 0)
		removeParticipant(db, ret);
	    return 0;
	default:
    }


    switch(pctohs(fromClient.opCode)) {
	case CMD_CONNECT_REQ:
	    if(msgLength > 4) {
		/* Redundant connect req */
		return 0;
	    }
	    sendConnectionReply(db, fd,	net_config, &client, 
				CAP_LITTLE_ENDIAN |
				ntohl(fromClient.connectReq.capabilities),
				ntohl(fromClient.connectReq.rcvbuf));
	    return 0;
	case CMD_GO:
	    doTransfer(fd, db, disk_config, net_config);
	    return 1;
	case CMD_DISCONNECT:
	    ret = lookupParticipant(db, &client);
	    if (ret >= 0)
		removeParticipant(db, ret);
	    return 0;
	default:
    }

    flprintf("Unexpected command %04x\n",(unsigned short) fromClient.opCode);

    return 0;
}

int startSender(struct disk_config *disk_config,
		struct net_config *net_config)
{
    char ipBuffer[16];
    int tries;

    participantsDb_t db;

    /* make the socket and print banner */
    int socket = makeSocket(SENDER_PORT(net_config->portBase));

    if(net_config->flags & FLAG_AUTORATE) {
	int q = getCurrentQueueLength(socket);
	if(q == 0) {
	    net_config->dir = 0;
	    net_config->sendbuf = getSendBuf(socket);
	} else {
	    net_config->dir = 1;
	    net_config->sendbuf = q;
	}
    }


    if(net_config->ttl == 1 && net_config->mcastAll == NULL) {
	getBroadCastAddress(socket, 
			    net_config->net_if,
			    &net_config->controlMcastAddr,
			    RECEIVER_PORT(net_config->portBase));
	setSocketToBroadcast(socket);
    } else {
	getMcastAllAddress(socket, 
			   net_config->net_if,
			   &net_config->controlMcastAddr,
			   net_config->mcastAll,
			   RECEIVER_PORT(net_config->portBase));
	setMcastDestination(socket, net_config->net_if,
			    &net_config->controlMcastAddr);
	setTtl(socket, net_config->ttl);
    }


    if(!(net_config->flags & FLAG_POINTOPOINT) &&
       ipIsZero(&net_config->dataMcastAddr)) {
	char ipBuffer[16];
	getDefaultMcastAddress(socket, net_config->net_if, 
			       &net_config->dataMcastAddr);
	flprintf("Using multicast address %s\n", 
        getIpString(&net_config->dataMcastAddr, ipBuffer));
	logprintf(log, "Using multicast address %s\n", 
        getIpString(&net_config->dataMcastAddr, ipBuffer));
	setMcastDestination(socket, net_config->net_if, 
			    &net_config->dataMcastAddr);
    }

    if(net_config->flags & FLAG_POINTOPOINT) {
	clearIp(&net_config->dataMcastAddr);
    }

    setPort(&net_config->dataMcastAddr, RECEIVER_PORT(net_config->portBase));

    flprintf("%sUDP sender for %s at ", 
	     disk_config->pipeName == NULL ? "" : "Compressed ",
	     disk_config->fileName == NULL ? "(stdin)" :
	     disk_config->fileName);
    printMyIp(net_config->net_if, socket);
    flprintf(" on %s \n", net_config->net_if);
    flprintf("Broadcasting control to %s\n",
	     getIpString(&net_config->controlMcastAddr, ipBuffer));
	logprintf(log, "Broadcasting control to %s\n",
	     getIpString(&net_config->controlMcastAddr, ipBuffer));

    net_config->capabilities = SENDER_CAPABILITIES;
    if(net_config->flags & FLAG_ASYNC)
	net_config->capabilities |= CAP_ASYNC;

    sendHello(net_config, socket);
    db = makeParticipantsDb();
    tries = 0;
    while(!mainDispatcher(socket, db, disk_config, net_config,
			  disk_config->fileName != NULL, &tries)) {}
    free(db);
    return 0;
}

/*
 * Do the actual data transfer
 */
static int doTransfer(int socket, 
		      participantsDb_t db,
		      struct disk_config *disk_config,
		      struct net_config *net_config)
{
    int i;
    int ret;
    int endianness=PC_ENDIAN;
    struct fifo fifo;
    sender_stats_t stats;
    int in;
    int origIn;
    int pid;
    int isPtP = isPointToPoint(db, net_config->flags);
    time_t stTime = time(0);

    if((net_config->flags & FLAG_POINTOPOINT) &&
       nrParticipants(db) != 1) {
	fatal(1,"pointopoint mode set, and %d participants instead of 1\n",
	      nrParticipants(db));
    }

    net_config->rcvbuf=0;

    for(i=0; i<MAX_CLIENTS; i++)
	if(isParticipantValid(db, i)) {
	    int pRcvBuf = getParticipantRcvBuf(db, i);
	    if(isPtP)
		copyIpFrom(&net_config->dataMcastAddr, getParticipantIp(db,i));
	    net_config->capabilities &= getParticipantCapabilities(db, i);
	    if(pRcvBuf != 0 && 
	       (net_config->rcvbuf == 0 || net_config->rcvbuf > pRcvBuf))
		net_config->rcvbuf = pRcvBuf;
	}


    flprintf("%s", ctime(&stTime));
    flprintf("Starting transfer: %08x\n", net_config->capabilities);

    if(net_config->capabilities & CAP_BIG_ENDIAN)
	endianness = NET_ENDIAN;

    if(! (net_config->capabilities & CAP_NEW_GEN)) {
       net_config->dataMcastAddr = net_config->controlMcastAddr;
       net_config->flags &= ~(FLAG_SN | FLAG_ASYNC);
    }
    if(net_config->flags & FLAG_BCAST)
       net_config->dataMcastAddr = net_config->controlMcastAddr;

    origIn = openFile(disk_config);
    stats = allocSenderStats(origIn);
    in = openPipe(disk_config, origIn, &pid);
    initFifo(&fifo, net_config->blockSize);
    ret = spawnNetSender(&fifo, socket, endianness, net_config, db, stats);
    localReader(disk_config, &fifo, in);

    /* if we have a pipe, now wait for that too */
    if(pid) {
	waitForProcess(pid, "Pipe");
    }

    pthread_join(fifo.thread, NULL);    
    flprintf("Transfer complete.\007\n");

    /* remove all participants */
    for(i=0; i < MAX_CLIENTS; i++) {
	removeParticipant(db, i);
    }
    flprintf("\n");
  
    return 0;
}
